from typing import Optional

import torch
from torch import nn


class InputNorm(nn.Module):
    """
    Computes a running mean/variance of input features and performs normalization.
    Adapted from https://www.johndcook.com/blog/standard_deviation/
    """

    # Pretend that `count` is a float to make MyPy happy
    count: float

    def __init__(self, num_features: int, cliprange: float = 5) -> None:
        super().__init__()

        self.cliprange = cliprange
        self.register_buffer("count", torch.tensor(0.0))
        self.register_buffer("mean", torch.zeros(num_features))
        self.register_buffer("squares_sum", torch.zeros(num_features))
        self.fp16 = False
        self._stddev: Optional[torch.Tensor] = None
        self._dirty = True
        self._frozen = False

    def freeze(self) -> None:
        """
        Freeze the running statistics, thus the normalization.
        """
        self._frozen = True

    def unfreeze(self) -> None:
        """
        Unfreeze the running statistics, thus the normalization.
        """
        self._frozen = False

    def update(self, input: torch.Tensor) -> None:
        self._dirty = True
        count = input.numel() // input.size(-1)
        if count == 0:
            return
        dreduce = tuple(range(0, input.dim() - 1))
        mean = input.mean(dim=dreduce)
        square_sum = ((input - mean) * (input - mean)).sum(dim=dreduce)
        if self.count == 0:
            self.count += count
            self.mean = mean
            self.squares_sum = square_sum
        else:
            # This does not follow directly Welford's method since it is a batched update
            # Instead we consider computing the statistics of two sets, A="current set so far" B="current batch"
            # See Chan, Tony F.; Golub, Gene H.; LeVeque, Randall J. (1979), "Updating Formulae and a Pairwise Algorithm for Computing Sample Variances.", Technical Report STAN-CS-79-773, Department of Computer Science, Stanford University.
            delta = mean - self.mean
            self.mean += delta * count / (count + self.count)
            self.squares_sum += square_sum + torch.square(
                delta
            ) * count * self.count / (count + self.count)
            self.count += count

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        with torch.no_grad():
            if self.training and not self._frozen:
                self.update(input)
            if self.count > 1:
                input = (input - self.mean) / self.stddev()
            input = torch.clamp(input, -self.cliprange, self.cliprange)

        return input.half() if self.fp16 else input

    def enable_fp16(self) -> None:
        # Convert buffers back to fp32, fp16 has insufficient precision and runs into overflow on squares_sum
        self.float()
        self.fp16 = True

    def stddev(self) -> torch.Tensor:
        if self._dirty or self._stddev is None:
            sd = torch.sqrt(self.squares_sum / (self.count - 1))
            sd[sd == 0] = 1
            self._stddev = sd
            self._dirty = False
        return self._stddev



from copy import deepcopy
from dataclasses import dataclass
from typing import Dict, List, Optional

import numpy as np
import ragged_buffer
from ragged_buffer import RaggedBufferF32

from entity_gym.env import ObsSpace


@dataclass
class TranslationConfig:
    """Settings for translating position features.

    Attributes:
        reference_entity: Entity type of the entity which will be placed at the origin.
        position_features: Names of position features used for translation.
        rotation_vec_features: Names of that gives the direction of the reference entity in radians. All entities are rotated by this value.
        rotation_angle_feature: Name of feature that gives the direction of the reference entity in radians. All entities are rotated by this value.
        add_dist_feature: Adds a feature that is the distance to the reference entity.
    """

    reference_entity: str
    position_features: List[str]
    rotation_vec_features: Optional[List[str]] = None
    rotation_angle_feature: Optional[str] = None
    add_dist_feature: bool = False

    def __post_init__(self) -> None:
        assert (
            self.rotation_vec_features is None or self.rotation_angle_feature is None
        ), "Only one of rotation_vec_features and rotation_angle_feature can be specified"


class TranslatePositions(TranslationConfig):
    def __init__(
        self,
        cfg: TranslationConfig,
        obs_space: ObsSpace,
    ):
        super().__init__(**cfg.__dict__)
        self.feature_indices = {
            entity_name: [
                entity.features.index(feature_name)
                for feature_name in cfg.position_features
            ]
            for entity_name, entity in obs_space.entities.items()
            if entity_name != cfg.reference_entity
            and all(
                [
                    feature_name in entity.features
                    for feature_name in cfg.position_features
                ]
            )
        }
        self.reference_indices = [
            obs_space.entities[cfg.reference_entity].features.index(feature_name)
            for feature_name in cfg.position_features
        ]
        self.orientation_vec_indices = (
            [
                obs_space.entities[cfg.reference_entity].features.index(feature_name)
                for feature_name in cfg.rotation_vec_features
            ]
            if cfg.rotation_vec_features is not None
            else None
        )
        self.orientation_angle_index = (
            obs_space.entities[cfg.reference_entity].features.index(
                cfg.rotation_angle_feature
            )
            if cfg.rotation_angle_feature is not None
            else None
        )
        self.reference_entity = cfg.reference_entity
        self.add_dist_feature = cfg.add_dist_feature

    def apply(self, entities: Dict[str, RaggedBufferF32]) -> None:
        if self.reference_entity not in entities:
            return
        reference_entity = entities[self.reference_entity]
        origin = reference_entity[:, :, self.reference_indices]
        if self.orientation_vec_indices is not None:
            orientation: Optional[RaggedBufferF32] = reference_entity[
                :, :, self.orientation_vec_indices
            ]
        elif self.orientation_angle_index is not None:
            angle = reference_entity[:, :, self.orientation_angle_index].as_array()
            orientation = RaggedBufferF32.from_array(
                np.hstack([np.cos(angle), np.sin(angle)]).reshape(-1, 1, 2)
            )
            # TODO: ragged_buffer.translate_rotate assumes that all input arguments are views, so apply identity view
            orientation = orientation[:, :, :]
        else:
            orientation = None
        for entity_name, indices in self.feature_indices.items():
            if entity_name in entities:
                if orientation is not None:
                    ragged_buffer.translate_rotate(
                        entities[entity_name][:, :, indices],
                        origin,
                        orientation,
                    )
                else:
                    feats = entities[entity_name][:, :, indices]
                    feats -= origin

                if self.add_dist_feature:
                    # TODO: efficiency
                    ea = entities[entity_name].as_array()
                    np.linalg.norm(ea[:, indices], axis=1).reshape(-1, 1)
                    entities[entity_name] = RaggedBufferF32.from_flattened(
                        np.concatenate(
                            [ea, np.linalg.norm(ea[:, indices], axis=1).reshape(-1, 1)],
                            axis=1,
                        ),
                        entities[entity_name].size1(),
                    )

    def transform_obs_space(self, obs_space: ObsSpace) -> ObsSpace:
        if self.add_dist_feature:
            obs_space = deepcopy(obs_space)
            for entity_name in self.feature_indices.keys():
                features = list(obs_space.entities[entity_name].features)
                features.append("TranslatePositions.distance")
                obs_space.entities[entity_name].features = features
        return obs_space



import numpy as np
import torch

from rogue_net.input_norm import InputNorm


def test_correct_normalization() -> None:
    generator = torch.random.manual_seed(0)
    np.random.seed(0)

    size = (1000, 12)
    examples_per_method = 10
    methods = [torch.rand, torch.randn]

    min_batch_size = 16
    max_batch_size = 128

    for method in methods:
        for _ in range(examples_per_method):
            sample = method(size, generator=generator)
            layer = InputNorm(size[1])
            remaining = size[0]
            while remaining > 0:
                # Select a number of samples to take
                batch_size = min(
                    remaining, np.random.randint(min_batch_size, max_batch_size)
                )
                # Take it
                batch = sample[size[0] - remaining : size[0] - remaining + batch_size]
                samples_seen_so_far = sample[: size[0] - remaining + batch_size]
                remaining -= batch_size
                # Compute output
                layer(batch)
                # Compute statistics on the samples seen so far
                mean_so_far = samples_seen_so_far.mean(dim=0)
                std_so_far = samples_seen_so_far.std(dim=0)

                assert layer.count == size[0] - remaining
                # Note that this does not work when atol=1e-7
                assert torch.allclose(layer.mean, mean_so_far, rtol=0, atol=1e-6)
                assert torch.allclose(layer.stddev(), std_so_far, atol=1e-6, rtol=0)


def test_cliprange() -> None:
    generator = torch.random.manual_seed(0)
    np.random.seed(0)

    size = (1000, 12)
    examples_per_method = 10
    methods = [torch.rand, torch.randn]

    min_batch_size = 16
    max_batch_size = 128

    for method in methods:
        for _ in range(examples_per_method):
            sample = method(size, generator=generator)
            cliprange = np.random.uniform() * 7
            layer = InputNorm(size[1], cliprange=cliprange)
            remaining = size[0]
            while remaining > 0:
                # Select a number of samples to take
                batch_size = min(
                    remaining, np.random.randint(min_batch_size, max_batch_size)
                )
                # Take it
                batch = sample[size[0] - remaining : size[0] - remaining + batch_size]
                remaining -= batch_size
                # Compute output
                out = layer(batch)
                assert torch.all(out >= -cliprange)
                assert torch.all(out <= cliprange)


def test_freeze() -> None:
    generator = torch.random.manual_seed(0)
    np.random.seed(0)

    size = (1000, 12)
    examples_per_method = 10
    methods = [torch.rand, torch.randn]

    min_batch_size = 16
    max_batch_size = 128

    for method in methods:
        for _ in range(examples_per_method):
            sample = method(size, generator=generator)
            cliprange = np.random.uniform() * 7
            layer = InputNorm(size[1], cliprange=cliprange)
            remaining = size[0]
            while remaining > 0:
                # Select a number of samples to take
                batch_size = min(
                    remaining, np.random.randint(min_batch_size, max_batch_size)
                )
                # Take it
                batch = sample[size[0] - remaining : size[0] - remaining + batch_size]
                remaining -= batch_size
                # Compute output
                out = layer(batch)
                layer.freeze()
                other = layer(batch)
                assert torch.allclose(other, out, atol=0)
                layer.unfreeze()



from dataclasses import dataclass

import torch


@dataclass
class RaggedTensor:
    """
    RaggedTensor represents a 3D tensor with a variable length sequence dimension.

    Example:
    entities in obs 0: [A0, A0, A0, B0]
    entities in obs 1: [A1, B1, B1]
    entities in obs 2: [A2, A2]

    data = [A0, A0, A0, B0, A1, B1, B1, A2, A2]
    batch_index = [0, 0, 0, 0, 1, 1, 1, 2, 2]
    lengths = [4, 3, 2]
    """

    data: torch.Tensor
    """Flattened tensor combining batch and sequence dimension of shape (entities, features)"""
    batch_index: torch.Tensor
    """Gives the batch index of each entity in `data`"""
    lengths: torch.Tensor
    """Gives the number of entities in each observation"""



import numpy as np
import torch
from ragged_buffer import RaggedBufferI64

from entity_gym.env import VecCategoricalActionMask
from rogue_net.categorical_action_head import CategoricalActionHead
from rogue_net.ragged_tensor import RaggedTensor


def test_empty_actors() -> None:
    head = CategoricalActionHead(d_model=4, n_choice=2)
    x = RaggedTensor(
        data=torch.zeros(12, 4),
        batch_index=torch.tensor([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 3, 3]),
        lengths=torch.tensor([4, 4, 2, 2]),
    )

    action, lengths, logprob, entropy, logits = head.forward(
        x,
        index_offsets=RaggedBufferI64.from_array(
            np.array([[[0]], [[4]], [[6]], [[8]]])
        ),
        mask=VecCategoricalActionMask(
            actors=RaggedBufferI64.from_flattened(
                np.zeros((0, 1), dtype=np.int64),
                lengths=np.array([0, 0, 0, 0], dtype=np.int64),
            ),
            mask=None,
        ),
        prev_actions=None,
    )
    assert action.shape == (0,)
    assert np.array_equal(lengths, np.array([0, 0, 0, 0]))
    assert logprob.shape == (0,)
    assert entropy.shape == (0,)
    assert logits.shape == (0, 2)
